//Copyright 2017 Looking Glass Factory Inc.
//All rights reserved.
//Unauthorized copying or distribution of this file, and the source code contained herein, is strictly prohibited.

function HoloPlay(scene, camera, renderer, focalPointVector, constantCenter, hiResRender){
    //This makes sure we don't try to render before initializing
    var intialized = false;
    
    //This shouldn't be necessary, but something isn't getting initialized immediately, so we wait a frame
    var shaderDelay = true;
    
    //Stores the distance to the focal plane
    //Let's us change the rotation or position of the camera and still have it work
    //Change this in order to change the focus of the camera after runtime
    var viewScale;
    var center;
            
    function init()
    {
        this.threeD = true;
        this.jsonObj = null;
        
        this.tilesX = 4;
        this.tilesY = 8;
        
        if(hiResRender === undefined){
            hiResRender = false;
        }
        
        if(focalPointVector === undefined){
            var vector = new THREE.Vector3();
            camera.getWorldDirection(vector); //Sets the vector to the camera forward
            
            viewScale = Math.max(camera.position.length(), 1); //Sets the focal distance to either the distance to center or just ahead of the camera
            //Because no center was provided in the constructor, it assumes that the center is 0,0,0
            center = new THREE.Vector3(0,0,0);
            
            vector.multiplyScalar(viewScale);
            focalPointVector = [camera.position.x + vector.x, camera.position.y + vector.y, camera.position.z + vector.z]; //Sets the focal point to the front of the camera as far away as it is from (0,0,0)
            
        } else{
            if(focalPointVector instanceof THREE.Vector3){
                focalPointVector = [focalPointVector.x, focalPointVector.y, focalPointVector.z];
            }
            center = new THREE.Vector3(focalPointVector[0], focalPointVector[1], focalPointVector[2]);
            viewScale = Math.max(center.distanceTo(camera.position), 1) //Sets the focal distance to either the distance to center or just ahead of the camera
        }
        if(constantCenter === undefined)
            constantCenter = true;
        
        this._renderer = renderer;
        this._camera = camera;
        this._scene = scene;
            
        //Locks the center to a fixed position if true, which is the default
        //Good for orbit controls, but should be false for things like first-person controls
        this.holdCenter = constantCenter;
        
        this.cameraForward = new THREE.Vector3();
        camera.getWorldDirection(cameraForward);
        
        //Capture settings
        this.numViews = this.tilesX * this.tilesY;
        this.viewCone = 40;

        //Init intermediate render textures
        this.smallRenderTargets = new Array(numViews);
        this.planes = new Array(numViews);
        this.planeMats = new Array(numViews);
        
        this.startNear = camera.near;
        this.startFar = camera.far;
        this.startDist = viewScale;
        
        //Buffer scene
        this.bufferScene = new THREE.Scene();
        this.bufferCam = new THREE.OrthographicCamera(-window.innerWidth/2, window.innerWidth/2, window.innerHeight/2, -window.innerHeight/2, 1, 3);
        bufferCam.position.z = 2;
        bufferScene.add(bufferCam);
        var renderResolution = 2048;
        if(hiResRender){
            renderResolution = 4096;
        }
        this.bufferSceneRender = new THREE.WebGLRenderTarget(renderResolution, renderResolution, {format: THREE.RGBFormat});
        
        //render texture dimensions
        this.renderSizeX = renderResolution / tilesX;
        this.renderSizeY = renderResolution / tilesY;
        
        //Init shader uniforms
        var uniforms = 
        {
            quiltTexture: {value: bufferSceneRender.texture},
            pitch: {value:0},
            tilt: {value:0},
            center: {value:0},
            invView: {value:0},
            flipX: {value:0},
            flipY: {value:0},
            subp: {value:0},
            ri: {value:0},
            bi: {value:2},
            numViews: {value:0},
            tilesX: {value:0},
            tilesY: {value:0},
            windowW: {value:0},
            windowH: {value:0},
            windowX: {value:0},
            windowY: {value:0},
            screenW: {value:0},
            screenH: {value:0}
        };

        //Set up the shader
        var shaderProperties = {
            uniforms: uniforms,
            vertexShader: VertexShaderCode, 
            fragmentShader: FragmentShaderCode
        };

        //Apply the shader to the buffer material
        this.bufferMat = new THREE.ShaderMaterial(shaderProperties);

        //Set up the final render scene
        this.finalRenderScene = new THREE.Scene();
        this.renderPlaneGeometry = new THREE.PlaneGeometry(window.innerWidth, window.innerHeight);
        
        this.renderPlane = new THREE.Mesh(renderPlaneGeometry, bufferMat);
        finalRenderScene.add(renderPlane);
        
        this.finalRenderCamera = new THREE.OrthographicCamera(-window.innerWidth/2, window.innerWidth/2, window.innerHeight/2, -window.innerHeight/2, 1, 3);
        finalRenderCamera.position.z = 2;
        finalRenderScene.add(finalRenderCamera);
        
        //Initialize up the 32 planes for the buffer scene
        setBufferScene();
        
        //Add the user buttons
        setupFullScreen();
        setupFileSelectionButton();
                
        initialized = true;
    };
    
    //******HTML SETUP******//
    
    //Create the dom element for the fullscreen button
    function makeFullScreenButton(){
        var newHTML = 
            '<input type="button" style="margin:20px; position:fixed; top:0px; right:0px; z-index: 10000; height:50px; width:150px;" id="fullscreenButton" value="Enter Full Screen Mode"/>';
        
        var buttonDiv = document.createElement("div");
        
        buttonDiv.innerHTML = newHTML;
        
        buttonDiv.setAttribute("id", "fullscreen");
        
        document.body.appendChild(buttonDiv);
    };
        
    //Adding the functionality for the fullscreen button
    function setupFullScreen(){
        makeFullScreenButton();
        
        document.getElementById('fullscreen').addEventListener("click", function(){
            if(_renderer.domElement.requestFullscreen) {
                _renderer.domElement.requestFullscreen();
            } else if(_renderer.domElement.mozRequestFullScreen) {
                _renderer.domElement.mozRequestFullScreen();
            } else if(_renderer.domElement.webkitRequestFullscreen) {
                _renderer.domElement.webkitRequestFullscreen();
            } else if(_renderer.msRequestFullscreen) {
                _renderer.domElement.msRequestFullscreen();
            }
        });
    };
    
    //Setup the dom element for the file selection button
    function setupFileSelectionButton(){
        
        var showButton = true;
        
        var storedValue = localStorage['Config'];
        
        if(storedValue){
            jsonObj = JSON.parse(storedValue);
                        
            setShaderValues(jsonObj.DPI.value, jsonObj.pitch.value, jsonObj.slope.value, jsonObj.screenH.value, jsonObj.screenW.value, jsonObj.center.value, jsonObj.flipImageX.value, jsonObj.flipImageY.value);
            
            viewCone = jsonObj.viewCone.value;
            
            showButton = false;
        } else{
            //Set up the shader for the holoplayer config with default values - it will look bad, but render
            //This should maybe just render in 2D instead
            setShaderValues(338.0, 49.89741516113281, 5.622990608215332, 1600.0, 2560.0, -0.5296876430511475, 0, 0, 1);
        }
        
        var newHTML = 
            '<div id="jsonInput" style="margin:20px; position:fixed; top:50px; right:0px; z-index: 10000; height:50px; width:150;" >'
            + '<input type="button" id="loadJSON" value="Load HoloPlay Configuration" />'
            + '<input type="file" style="display:none;" id="file" name="file"/></div>';
        
        var fileButtonDiv = document.createElement("div");
        
        fileButtonDiv.innerHTML = newHTML;
        
        document.body.appendChild(fileButtonDiv);
        
        document.getElementById('file').addEventListener('change', handleFileSelect, false);
        
        //Syncing the functionality of the button to the hidden file object
        document.getElementById('loadJSON').addEventListener("click", function(){document.getElementById('file').click()});

        //Don't show the button if we already have a JSON cached
        //Need to have the html in the document though to have the reset work
        if(!showButton){
            fileButtonDiv.style = "display:none";
        }
    };
    
    //Add a listener to change the config file
    function handleFileSelect(evt) {
        var files = evt.target.files; // FileList object
        if(files[0].name !== "visual.json"){
            alert("Selected file is not a calibration file!");
            return;
        };

        var fileReader = new FileReader();
        fileReader.readAsText(files[0]);
        fileReader.onload = function(evt){  
            jsonObj = JSON.parse(evt.target.result);
                
            setShaderValues(jsonObj.DPI.value, jsonObj.pitch.value, jsonObj.slope.value, jsonObj.screenH.value, jsonObj.screenW.value, jsonObj.center.value, jsonObj.flipImageX.value, jsonObj.flipImageY.value);
            
            viewCone = jsonObj.viewCone.value;

            document.getElementById("jsonInput").setAttribute("STYLE", "display:none");// = "display:none"; 
            
            localStorage['Config'] = evt.target.result;
        };


    };
    
    //*******SHADER SETUP******//
    
    //Set up the buffer scene
    function setBufferScene(){
        //For future reference, see https://threejs.org/docs/index.html#api/renderers/WebGLRenderer.copyTextureToTexture
        var planeX = window.innerWidth/this.tilesX;
        var planeY = window.innerHeight/this.tilesY;
        var startX = -window.innerWidth/2 + planeX/2;
        var startY = -window.innerHeight/2 + planeY/2;
        var planeGeometry = new THREE.PlaneBufferGeometry(planeX, planeY);

        var planeIndex = 0;
        
        for(var i = 0; i < numViews; i++){
            planeMats[i] = new THREE.MeshBasicMaterial(0x000000);
            smallRenderTargets[i] = new THREE.WebGLRenderTarget(this.renderSizeX, this.renderSizeY, {format: THREE.RGBFormat});
            planeMats[i].map = smallRenderTargets[i].texture;
        }
        
        //We will always have 4 x 8 grid
        //even if numViews is less than 32
        for(var i = 0; i < this.tilesY; i++){
            for(var j = 0; j < this.tilesX; j++){
                if(planeIndex >= numViews)
                    break;
                planes[planeIndex] = new THREE.Mesh(planeGeometry, planeMats[planeIndex]);
                planes[planeIndex].position.set(startX + planeX * j, startY + planeY * i, 0);
                bufferScene.add(planes[planeIndex]);
                planeIndex++;
            };
        };
    };
    
    function setShaderValues(dpi, pitch, slope, screenH, screenW, center, flipX, flipY, invView)
    {
        var screenInches = screenW / dpi;
        var newPitch = pitch * screenInches;
        
        //account for tilt in measuring pitch horizontally
        newPitch *= Math.cos(Math.atan(1.0 / slope));
        bufferMat.uniforms.pitch.value = newPitch;
                
        //tilt
        var newTilt = screenH / (screenW * slope);
        if(flipX == 1)
            newTilt *= -1;
        bufferMat.uniforms.tilt.value = newTilt;
        
        //center
        bufferMat.uniforms.center.value = center;
        
        //screen and window values to offset the image
        bufferMat.uniforms.screenW.value = screenW;
        bufferMat.uniforms.screenH.value = screenH;
        bufferMat.uniforms.windowW.value = window.innerWidth;
        bufferMat.uniforms.windowH.value = window.innerHeight;
        bufferMat.uniforms.windowX.value = window.screenX;
        bufferMat.uniforms.windowY.value = window.screenY;
        
        //should we invert?
        bufferMat.uniforms.invView.value = invView;
        
        //Should we flip it for peppers?
        bufferMat.uniforms.flipX.value = flipX;
        bufferMat.uniforms.flipY.value = flipY;
        
        bufferMat.uniforms.subp.value = 1/(screenW * 3);

        //tiles
        bufferMat.uniforms.tilesX.value = this.tilesX;
        bufferMat.uniforms.tilesY.value = this.tilesY;
        
        bufferMat.needsUpdate = true;        
    };
    
    //*******LOGIC FOR CAPTURING MULTIPLE VIEWS******//

    //Render the different views
    function captureViews(scene, camera)
    {        
        var origPosition = new THREE.Vector3(camera.position.x, camera.position.y, camera.position.z);
        var start = -viewCone/2;
        var end = viewCone/2;
        var distance = center.distanceTo(camera.position);
        var size = 2 * distance * Math.tan(0.5 * THREE.Math.degToRad(camera.fov));
        
        for(var i = 0; i < numViews; i++)
        {
            camera.position.set(origPosition.x, origPosition.y, origPosition.z);
            var radians = THREE.Math.degToRad(THREE.Math.lerp(start, end, i/(numViews - 1)));
            
            handleOffset(camera, radians, distance, size);
            
            renderer.render(scene, camera, smallRenderTargets[i], true);
            //Shouldn't need this
            planeMats[i].needsUpdate = true;
        }
        camera.position.set(origPosition.x, origPosition.y, origPosition.z);
    };
    
    //Move the camera 
    function handleOffset(camera, angle, distance, size)
    {   
        //angle needs to be in radians
        var offsetX = distance * Math.tan(angle); 
        
        //Get the right direction                
        var tempRight = new THREE.Vector3(camera.right.x * offsetX, camera.right.y * offsetX, camera.right.z * offsetX);
        
        camera.position.add(tempRight);
            
        camera.projectionMatrix.elements[8] = -2 * offsetX / (size * camera.aspect);
    };
    
    function doFinalRender()
    {
        renderer.render(bufferScene, bufferCam, bufferSceneRender, true);
        
        //Shouldn't need this either
        bufferMat.map = bufferSceneRender.texture;
        bufferMat.needsUpdate = true;
    };

    //Render loop, with options for 3D or 2D rendering
    HoloPlay.prototype.render = function (scene, camera, renderer){
        if(!initialized)
            return;
        else if(shaderDelay){
            shaderDelay = false;
            setShaderValues(jsonObj.DPI.value, jsonObj.pitch.value, jsonObj.slope.value, jsonObj.screenH.value, jsonObj.screenW.value, jsonObj.center.value, jsonObj.flipImageX.value, jsonObj.flipImageY.value, jsonObj.invView.value);  
        }
        
        if(scene === undefined)
            scene = _scene;
        if(camera === undefined)
            camera = _camera;
        if(renderer === undefined)
            renderer = _renderer;
        
        if(!threeD){
            if(camera.projectionMatrix.elements[8] != 0)
                camera.projectionMatrix.elements[8] = 0;
            renderer.render(scene, camera);
        } else{
            var worldRight = new THREE.Vector3(1,0,0);
            camera.right = worldRight.applyQuaternion(camera.quaternion);        
            
            if(holdCenter === false){
                camera.getWorldDirection(cameraForward);
                cameraForward.multiplyScalar(viewScale);

                center.addVectors(camera.position, cameraForward);
            } else{
                var dist = camera.position.distanceTo(center);
                camera.near = startNear * dist / viewScale;
                camera.far = startFar * dist / viewScale;
                camera.updateProjectionMatrix ();
            }

            captureViews(scene, camera);
            doFinalRender();

            renderer.render(finalRenderScene, finalRenderCamera);
        }
    };
    
    //*****EVENT LISTENERS*****//
    
    //Reset shader values on window resize to make it draw properly
    window.addEventListener('resize', function(){
        setShaderValues(jsonObj.DPI.value, jsonObj.pitch.value, jsonObj.slope.value, jsonObj.screenH.value, jsonObj.screenW.value, jsonObj.center.value, jsonObj.flipImageX.value, jsonObj.flipImageY.value, jsonObj.invView.value);  
    });
    
    //Setting key listeners, Forward Slash for switching between 2D and 3D
    //Tilda to load a new calibration file
    document.addEventListener('keydown', function(event){ 
            if(event.keyCode === 220){ 
                threeD = !threeD;
            } 
            else if(event.which == 192)
            {
                document.getElementById('file').click();
            } 
        }, false);
    
    
    //SHADER CODE
    var VertexShaderCode =         
            "varying vec2 iUv;"+
        
            "void main() {"+        
                "iUv = uv;"+
                "vec4 modelViewPosition = modelViewMatrix * vec4(position, 1.0);"+
                "gl_Position = projectionMatrix * modelViewPosition;"+
            "}";
    
    var FragmentShaderCode = 
            "uniform sampler2D quiltTexture;"+
            "uniform float pitch;"+
            "uniform float tilt;"+
            "uniform float center;"+
            "uniform float invView;" +
            "uniform float flipX;" +
            "uniform float flipY;" +
            "uniform float subp;" +
            "uniform int ri;"+
            "uniform int bi;"+
            "uniform float tilesX;"+ 
            "uniform float tilesY;"+
            "uniform float windowW;"+
            "uniform float windowH;"+
            "uniform float windowX;"+
            "uniform float windowY;"+
            "uniform float screenW;"+
            "uniform float screenH;"+
            "varying vec2 iUv;"+

            "vec2 texArr(vec3 uvz) {"+
                "float z = floor(uvz.z * tilesX * tilesY);"+
                "float x = (mod(z, tilesX) + uvz.x) / tilesX;"+
                "float y = (floor(z / tilesX) + uvz.y) / tilesY;"+
//                "x = saturate(x);"+
//                "y = saturate(y);"+
                "return vec2(x, y);"+
            "}"+

            "void main()"+
            "{"+
                "vec4 rgb[3];"+
                "vec3 nuv = vec3(iUv.xy, 0.0);"+
                
                //Set the values based off the position on the screen
//                "nuv.x *= windowW;"+
//                "nuv.x += windowX;"+
//                "nuv.x /= screenW;"+
//                "nuv.y *= windowH;"+
//                "nuv.y += screenH - windowY - windowH;"+
//                "nuv.y /= screenH;"+
        
                //Flip UVs if necessary
                "nuv.x = (1.0 - flipX) * nuv.x + flipX * (1.0 - nuv.x);"+
                "nuv.y = (1.0 - flipY) * nuv.y + flipY * (1.0 - nuv.y);"+
                
        
                "for (int i = 0; i < 3; i++) {"+
                    "nuv.z = (nuv.x + float(i) * subp + iUv.y * tilt) * pitch - center;"+
                    "nuv.z = mod(nuv.z + ceil(abs(nuv.z)), 1.0);"+
                    "nuv.z = (1.0 - invView) * nuv.z + invView * (1.0 - nuv.z);" +
                    "rgb[i] = texture2D(quiltTexture, texArr(nuv));"+
                "}"+
                
//                "gl_FragColor = vec4(rgb[ri].r, rgb[1].g, rgb[bi].b, 1);"+
                "gl_FragColor = vec4(rgb[0].r, rgb[1].g, rgb[2].b, 1);"+
            "}"
        ;

    //Call our initialization function once all our values are set
    init();
}